# -*- coding: utf-8 -*-
"""
Created on Thu Feb  6 12:38:05 2020

@author: keelin
"""


class BaseAlgorithm:
    """
    In case we want common functionality for all algorithms
    """

    def __init__(self):
        pass  # stuff to run on instance creation goes here
